/*** HDRCompression.cpp ***/

#include "MixedDomainFilter.h"
#include "..\\HDRLoader\\HDRLoader.h"
#include <cv.h>
#pragma comment(lib, "cv.lib")
#include <cxcore.h>
#pragma comment(lib, "cxcore.lib")
#include <highgui.h>
#pragma comment(lib, "highgui.lib")
#include <iostream>
using std::swap;
using std::min;
using std::max;

double fastSel(double *x, int p, int q, int tar) {
	if (p == q) return x[tar];
    int i = p, j = q;
	double md = x[(p+q)>>1];
    while (i <= j) {
        while (x[i] < md) i++;
        while (x[j] > md) j--;
        if (i > j) break;
        swap(x[i++], x[j--]);
    }
	if (j < tar && tar < i) return x[tar];
	if (tar <= j) return fastSel(x, p, j, tar);
    if (i <= tar) return fastSel(x, i, q, tar);
	return 0.00;
}
void findMaxMin(int W, int H, double *x, double &xMax, double &xMin, double c) {
	double *a = new double[W*H];
	memcpy(a, x, W*H*sizeof(double));
	xMax = fastSel(a, 0, W*H-1, int(W*H*(1.00-c)-0.5));
	xMin = fastSel(a, 0, W*H-1, int(W*H*c-0.5));
	delete []a;
}
extern inline void rangeClamp(double &x, double a, double b);
void HDRCompression(string xFileName, string yFileName, double alpha, double beta, double sigma, double lambda, int level, double percentile) {
	std::cout << ">> HDR Compression: " << yFileName.c_str() << "..." << std::endl;

	HDRLoaderResult HDRImage;
	HDRLoader::load(xFileName.c_str(), HDRImage);

	int W = HDRImage.width;
	int H = HDRImage.height;
	IplImage *xImage = cvCreateImage(cvSize(W, H), 8, 3);

	double *x = new double[W*H];
	double *y = new double[W*H];
	for(int index = 0;  index < W*H;  index++) {
		double L = (20*HDRImage.cols[3*index]+40*HDRImage.cols[3*index+1]+1*HDRImage.cols[3*index+2])/61;
		L = max(0.000001, L);
		for(int k = 0;  k < 3;  k++) {
			HDRImage.cols[3*index+k] /= float(L);
		}
		x[index] = log(L);
	}

	pyramidFilter(W, H, x, y, lambda, alpha, beta, sigma, false, false, level, 1.00);

	double yMax, yMin;
	findMaxMin(W, H, y, yMax, yMin, percentile/100);
	double gamma = log(100.0)/(yMax-yMin);
	for(int i = 0;  i < W;  i++) {
		for(int j = 0;  j < H;  j++) {
			double L = exp(gamma*(y[j*W+i]-yMax));
			for(int k = 0;  k < 3;  k++) {
				double color = L*double(HDRImage.cols[3*(j*W+i)+k]);
				rangeClamp(color, 0.00, 1.00);
				color = pow(color, 1.00/2.20);
				xImage->imageData[j*xImage->widthStep+3*i+(2-k)] = char(int(255*color+0.50));
			}
		}
	}
	cvSaveImage(yFileName.c_str(), xImage);

	cvReleaseImage(&xImage);
	delete []x;
	delete []y;
}

void toneAdjustment(string xFileName, string yFileName, double alpha, double beta, double sigma, double lambda, int level, double percentile) {
	std::cout << ">> Tone Adjustment: " << yFileName.c_str() << "..." << std::endl;

	IplImage *xImage = cvLoadImage(xFileName.c_str(), 1);
	IplImage *yImage = cvCloneImage(xImage);
	int W = xImage->width;
	int H = xImage->height;

	double *x = new double[W*H];
	double *y = new double[W*H];
	double *L = new double[W*H];
	for(int i = 0;  i < W;  i++) {
		for(int j = 0;  j < H;  j++) {
			double r = double(unsigned char(xImage->imageData[j*xImage->widthStep+3*i+2]))/255.00;
			double g = double(unsigned char(xImage->imageData[j*xImage->widthStep+3*i+1]))/255.00;
			double b = double(unsigned char(xImage->imageData[j*xImage->widthStep+3*i+0]))/255.00;
			L[j*W+i] = (r+g+b)/3;
			L[j*W+i] = max(0.000001, L[j*W+i]);
			x[j*W+i] = log(L[j*W+i]);
		}
	}

	pyramidFilter(W, H, x, y, lambda, alpha, beta, sigma, false, false, level, 1.00);

	double yMax, yMin;
	findMaxMin(W, H, y, yMax, yMin, percentile/100);
	double gamma = log(100.0)/(yMax-yMin);
	for(int i = 0;  i < W;  i++) {
		for(int j = 0;  j < H;  j++) {
			double ratioL = exp(gamma*(y[j*W+i]-yMax))/L[j*W+i];
			if(L[j*W+i] < 5.00/255) {
				ratioL = 1.00;
			}else if(L[j*W+i] < 10.00/255) {
				double alpha = L[j*W+i]*255/5.00-1;
				ratioL = 1.00-alpha*(1.00-exp(gamma*(y[j*W+i]-yMax))/L[j*W+i]);
			}
			for(int k = 0;  k < 3;  k++) {
				double color = ratioL*double(unsigned char(xImage->imageData[j*xImage->widthStep+3*i+k]))/255.00;
				rangeClamp(color, 0.00, 1.00);
				color = pow(color, 1.00/2.20);
				yImage->imageData[j*xImage->widthStep+3*i+k] = char(int(255*color+0.50));
			}
		}
	}
	cvSaveImage(yFileName.c_str(), yImage);

	cvReleaseImage(&xImage);
	cvReleaseImage(&yImage);
	delete []x;
	delete []y;
	delete []L;
}
