/*** MixedDomainFilter.cpp ***/

#include "MixedDomainFilter.h"
#include "..//FFTW//fftw3.h"
#pragma comment(lib, "..//FFTW//libfftw3-3.lib")
#include <cv.h>
#pragma comment(lib, "cv.lib")
#include <cxcore.h>
#pragma comment(lib, "cxcore.lib")
#include <highgui.h>
#pragma comment(lib, "highgui.lib")
#include <iostream>
#include <cmath>
using std::min;
using std::max;

void linearFilter(int W, int H, double *x, double *y, double beta) {
	double xSum = 0.00;
	for(int index = 0;  index < W*H;  index++) {
		xSum += x[index];
	}
	double xAverage = xSum/(W*H);
	for(int index = 0;  index < W*H;  index++) {
		y[index] = xAverage+beta*(x[index]-xAverage);
	}
}

inline void rangeClamp(double &x, double a, double b) {
	x = max(min(x, b), a);
}
inline int symmetricClamp(int x, int a, int b, int aa = 0, int bb = 0) {
	if(x < a) {
		return 2*a-x-aa;
	}else if(x > b) {
		return 2*b-x+bb;
	}else {
		return x;
	}
}
void MixedDomainFilter(int W, int H, double *x, double *y, double *b, double lambda, double alpha, double beta, double sigma, bool useTable, double range) {
	double *L = new double[W*H];
	double *FL = new double[W*H];
	double *Ki = new double[W];
	double *Kj = new double[H];
	double *FKi = new double[W];
	double *FKj = new double[H];

	int M = 0;
	for(int i = 1;  i < W;  i++) {
		for(int j = 1;  j < H;  j++) {
			if(i*i+j*j <= N*N) {
				M++;
			}
		}
	}
	M = 4*M+4*N+1;
#pragma omp parallel for
	for(int i = 0;  i < W;  i++) {
#pragma omp parallel for
		for(int j = 0;  j < H;  j++) {
			if(i*i+j*j <= N*N) {
				L[j*W+i] = 1.00/M;
			}else {
				L[j*W+i] = 0.00;
			}
		}
	}
#pragma omp parallel for
	for(int i = 0;  i < W;  i++) {
		if(i <= NK) {
			Ki[i] = coeffK[i];
		}else {
			Ki[i] = 0.00;
		}
	}
#pragma omp parallel for
	for(int j = 0;  j < H;  j++) {
		if(j <= NK) {
			Kj[j] = coeffK[j];
		}else {
			Kj[j] = 0.00;
		}
	}

	double *s = new double[W*H];
	double *t = new double[W*H];
	double *Fy = new double[W*H];
#pragma omp parallel for
	for(int index = 0;  index < W*H;  index++) {
		s[index] = 0.00;
		t[index] = 0.00;
	}
#pragma omp parallel for
	for(int i = -N;  i < W+N;  i++) {
#pragma omp parallel for
		for(int j = -N;  j < H+N;  j++) {
			localFilter(W, H, x, i, j, s, t, alpha, beta, sigma, useTable, range);
		}
	}
#pragma omp parallel for
	for(int index = 0;  index < W*H;  index++) {
		s[index] /= M;
		t[index] /= M;
	}
	// compute: Fy = lambda*FK*Fb+Ft-FL*Fs
#pragma omp parallel for
	for(int i = 0;  i < W;  i++) {
#pragma omp parallel for
		for(int j = 0;  j < H;  j++) {
			y[j*W+i] = 0.00;
			for(int ii = i-NK;  ii <= i+NK;  ii++) {
				int ii_ = symmetricClamp(ii, 0, W-1);
				y[j*W+i] += b[j*W+ii_]*coeffK[abs(ii-i)];
			}
		}
	}
#pragma omp parallel for
	for(int i = 0;  i < W;  i++) {
#pragma omp parallel for
		for(int j = 0;  j < H;  j++) {
			b[j*W+i] = 0.00;
			for(int jj = j-NK;  jj <= j+NK;  jj++) {
				int jj_ = symmetricClamp(jj, 0, H-1);
				b[j*W+i] += y[jj_*W+i]*coeffK[abs(jj-j)];
			}
		}
	}
#pragma omp parallel for
	for(int i = 0;  i < W;  i++) {
#pragma omp parallel for
		for(int j = 0;  j < H;  j++) {
			y[j*W+i] = 0.00;
			for(int ii = i-N;  ii <= i+N;  ii++) {
				for(int jj = j-N;  jj <= j+N;  jj++) {
					int ii_ = symmetricClamp(ii, 0, W-1);
					int jj_ = symmetricClamp(jj, 0, H-1);
					if((ii-i)*(ii-i)+(jj-j)*(jj-j) <= N*N) {
						y[j*W+i] += s[jj_*W+ii_];
					}
				}
			}
			y[j*W+i] /= M;
		}
	}
#pragma omp parallel for
	for(int index = 0;  index < W*H;  index++) {
		y[index] = lambda*b[index]+t[index]-y[index];
	}
	fftw_plan p;
	p = fftw_plan_r2r_1d(W, Ki, FKi, FFTW_REDFT00, FFTW_ESTIMATE);
	fftw_execute(p);
	p = fftw_plan_r2r_1d(H, Kj, FKj, FFTW_REDFT00, FFTW_ESTIMATE);
	fftw_execute(p);
	p = fftw_plan_r2r_2d(H, W, L, FL, FFTW_REDFT00, FFTW_REDFT00, FFTW_ESTIMATE);
	fftw_execute(p);
	p = fftw_plan_r2r_2d(H, W, y, Fy, FFTW_REDFT00, FFTW_REDFT00, FFTW_ESTIMATE);
	fftw_execute(p);
	fftw_destroy_plan(p);

#pragma omp parallel for
	for(int index = 0;  index < W*H;  index++) {
		int i = index%W;
		int j = index/W;
		double FK = FKi[i]*FKj[j];
		Fy[index] /= (lambda*FK*FK+1.00-FL[index]*FL[index]);
	}
	p = fftw_plan_r2r_2d(H, W, Fy, y, FFTW_REDFT00, FFTW_REDFT00, FFTW_ESTIMATE);
	fftw_execute(p);
	fftw_destroy_plan(p);

#pragma omp parallel for
	for(int index = 0;  index < W*H;  index++) {
		y[index] /= (4*(W-1)*(H-1));
	}
	delete []L;
	delete []FL;
	delete []Ki;
	delete []Kj;
	delete []FKi;
	delete []FKj;
	delete []s;
	delete []t;
	delete []Fy;
}
void jointMixedDomainFilter(int W, int H, double *x, double **ref, double *y, double *b, double lambda, double alpha, double beta, double sigma,
					bool useTable, double range) {
	double *L = new double[W*H];
	double *FL = new double[W*H];
	double *Ki = new double[W];
	double *Kj = new double[H];
	double *FKi = new double[W];
	double *FKj = new double[H];

	int M = 0;
	for(int i = 1;  i < W;  i++) {
		for(int j = 1;  j < H;  j++) {
			if(i*i+j*j <= N*N) {
				M++;
			}
		}
	}
	M = 4*M+4*N+1;
#pragma omp parallel for
	for(int i = 0;  i < W;  i++) {
#pragma omp parallel for
		for(int j = 0;  j < H;  j++) {
			if(i*i+j*j <= N*N) {
				L[j*W+i] = 1.00/M;
			}else {
				L[j*W+i] = 0.00;
			}
		}
	}
#pragma omp parallel for
	for(int i = 0;  i < W;  i++) {
		if(i <= NK) {
			Ki[i] = coeffK[i];
		}else {
			Ki[i] = 0.00;
		}
	}
#pragma omp parallel for
	for(int j = 0;  j < H;  j++) {
		if(j <= NK) {
			Kj[j] = coeffK[j];
		}else {
			Kj[j] = 0.00;
		}
	}

	double *s = new double[W*H];
	double *t = new double[W*H];
	double *Fy = new double[W*H];
#pragma omp parallel for
	for(int index = 0;  index < W*H;  index++) {
		s[index] = 0.00;
		t[index] = 0.00;
	}
#pragma omp parallel for
	for(int i = -N;  i < W+N;  i++) {
#pragma omp parallel for
		for(int j = -N;  j < H+N;  j++) {
			localFilter(W, H, x, i, j, s, t, alpha, beta, sigma, useTable, range);
		}
	}
#pragma omp parallel for
	for(int index = 0;  index < W*H;  index++) {
		s[index] /= M;
		t[index] /= M;
	}
	// compute: Fy = lambda*FK*Fb+Ft-FL*Fs
#pragma omp parallel for
	for(int i = 0;  i < W;  i++) {
#pragma omp parallel for
		for(int j = 0;  j < H;  j++) {
			y[j*W+i] = 0.00;
			for(int ii = i-NK;  ii <= i+NK;  ii++) {
				int ii_ = symmetricClamp(ii, 0, W-1);
				y[j*W+i] += b[j*W+ii_]*coeffK[abs(ii-i)];
			}
		}
	}
#pragma omp parallel for
	for(int i = 0;  i < W;  i++) {
#pragma omp parallel for
		for(int j = 0;  j < H;  j++) {
			b[j*W+i] = 0.00;
			for(int jj = j-NK;  jj <= j+NK;  jj++) {
				int jj_ = symmetricClamp(jj, 0, H-1);
				b[j*W+i] += y[jj_*W+i]*coeffK[abs(jj-j)];
			}
		}
	}
#pragma omp parallel for
	for(int i = 0;  i < W;  i++) {
#pragma omp parallel for
		for(int j = 0;  j < H;  j++) {
			y[j*W+i] = 0.00;
			for(int ii = i-N;  ii <= i+N;  ii++) {
				for(int jj = j-N;  jj <= j+N;  jj++) {
					int ii_ = symmetricClamp(ii, 0, W-1);
					int jj_ = symmetricClamp(jj, 0, H-1);
					if((ii-i)*(ii-i)+(jj-j)*(jj-j) <= N*N) {
						y[j*W+i] += s[jj_*W+ii_];
					}
				}
			}
			y[j*W+i] /= M;
		}
	}
#pragma omp parallel for
	for(int index = 0;  index < W*H;  index++) {
		y[index] = lambda*b[index]+t[index]-y[index];
	}
	fftw_plan p;
	p = fftw_plan_r2r_1d(W, Ki, FKi, FFTW_REDFT00, FFTW_ESTIMATE);
	fftw_execute(p);
	p = fftw_plan_r2r_1d(H, Kj, FKj, FFTW_REDFT00, FFTW_ESTIMATE);
	fftw_execute(p);
	p = fftw_plan_r2r_2d(H, W, L, FL, FFTW_REDFT00, FFTW_REDFT00, FFTW_ESTIMATE);
	fftw_execute(p);
	p = fftw_plan_r2r_2d(H, W, y, Fy, FFTW_REDFT00, FFTW_REDFT00, FFTW_ESTIMATE);
	fftw_execute(p);
	fftw_destroy_plan(p);

#pragma omp parallel for
	for(int index = 0;  index < W*H;  index++) {
		int i = index%W;
		int j = index/W;
		double FK = FKi[i]*FKj[j];
		Fy[index] /= (lambda*FK*FK+1.00-FL[index]*FL[index]);
	}
	p = fftw_plan_r2r_2d(H, W, Fy, y, FFTW_REDFT00, FFTW_REDFT00, FFTW_ESTIMATE);
	fftw_execute(p);
	fftw_destroy_plan(p);

#pragma omp parallel for
	for(int index = 0;  index < W*H;  index++) {
		y[index] /= (4*(W-1)*(H-1));
	}
	delete []L;
	delete []FL;
	delete []Ki;
	delete []Kj;
	delete []FKi;
	delete []FKj;
	delete []s;
	delete []t;
	delete []Fy;
}

inline double nonLinearMap(double d, double alpha, double beta, double sigma, double range) {
	double fd = d;
	double r = abs(d)/sigma;
	double dRef = (d > 0.00 ? sigma : -sigma);
	if(r > 1.00) {
		fd = dRef*sqrt(1.00+beta*beta*(r*r-1.00));
	}else if(alpha == 1.00) {
		fd = d;
	}else if(alpha < 1.00) {
		fd = dRef*pow(r, 1.00/alpha);
	}else {
		double s = max(0.00, min(1.00, 100*(r*sigma-0.01*range)));
		fd = dRef*(r+s*(pow(r, 1.00/alpha)-r));
	}
	return fd;
}
double nonLinearArray[256];
void computeTable(double alpha, double beta, double sigma, double range) {
	for(int index = 0;  index < 256;  index++) {
		nonLinearArray[index] = nonLinearMap(double(index)/255, alpha, beta, sigma, range);
	}
}

void localFilter(int W, int H, double *x, int i, int j, double *s, double *t, double alpha, double beta, double sigma, bool useTable, double range) {
	for(int ii = i-N;  ii <= i;  ii++) {
		for(int jj = j-N;  jj <= (ii == i ? j : j+N);  jj++) {
			if((ii-i)*(ii-i)+(jj-j)*(jj-j) > N*N) {
				continue;
			}
			int p = j*W+i;
			int q = jj*W+ii;
			bool pInside = (i >= 0 && i < W && j >= 0 && j < H);
			bool qInside = (ii >= 0 && ii < W && jj >= 0 && jj < H);
			if(pInside == false && qInside == false) {
				continue;
			}
			int pi = symmetricClamp(i, 0, W-1);
			int pj = symmetricClamp(j, 0, H-1);
			int qi = symmetricClamp(ii, 0, W-1);
			int qj = symmetricClamp(jj, 0, H-1);
			double xp = x[pj*W+pi];
			double xq = x[qj*W+qi];
			double d = xq-xp;
			double fd = d;
			if(useTable == true) {
				rangeClamp(d, -1.00, 1.00);
				if(d > 0.00) {
					fd = nonLinearArray[int(255*d+0.50)];
				}else {
					fd = -nonLinearArray[int(255*(-d)+0.50)];
				}
			}else {
				fd = nonLinearMap(d, alpha, beta, sigma, range);
			}
			double zpq = xp+fd;
			double zqp = xq-fd;
			if(pInside == true) {
				s[p] += zpq;
				t[p] += zqp;
			}
			if(qInside == true) {
				s[q] += zqp;
				t[q] += zpq;
			}
		}
	}
}
void jointLocalFilter(int W, int H, double *x, double **ref, int i, int j, double *s, double *t, double alpha, double beta, double sigma,
					  bool useTable, double range) {
	for(int ii = i-N;  ii <= i;  ii++) {
		for(int jj = j-N;  jj <= (ii == i ? j : j+N);  jj++) {
			if((ii-i)*(ii-i)+(jj-j)*(jj-j) > N*N) {
				continue;
			}
			int p = j*W+i;
			int q = jj*W+ii;
			bool pInside = (i >= 0 && i < W && j >= 0 && j < H);
			bool qInside = (ii >= 0 && ii < W && jj >= 0 && jj < H);
			if(pInside == false && qInside == false) {
				continue;
			}
			int pi = symmetricClamp(i, 0, W-1);
			int pj = symmetricClamp(j, 0, H-1);
			int qi = symmetricClamp(ii, 0, W-1);
			int qj = symmetricClamp(jj, 0, H-1);
			double xp = x[pj*W+pi];
			double xq = x[qj*W+qi];
			double d = xq-xp;
			double refD = sqrt(((ref[0][pj*W+pi]-ref[0][qj*W+qi])*(ref[0][pj*W+pi]-ref[0][qj*W+qi])+
				(ref[1][pj*W+pi]-ref[1][qj*W+qi])*(ref[1][pj*W+pi]-ref[1][qj*W+qi])+
				(ref[2][pj*W+pi]-ref[2][qj*W+qi])*(ref[2][pj*W+pi]-ref[2][qj*W+qi]))/3);
			double fd;
			if(refD == 0.00) {
				fd = 0.00;
			}else {
				fd = d*nonLinearMap(refD, alpha, beta, sigma, range)/refD;
			}
			double zpq = xp+fd;
			double zqp = xq-fd;
			if(pInside == true) {
				s[p] += zpq;
				t[p] += zqp;
			}
			if(qInside == true) {
				s[q] += zqp;
				t[q] += zpq;
			}
		}
	}
}
