/*** PyramidFilter.cpp ***/

#include "MixedDomainFilter.h"
#include "..//FFTW//fftw3.h"
#pragma comment(lib, "..//FFTW//libfftw3-3.lib")
#include <iostream>
#include <cmath>
using std::min;
using std::max;

extern inline int symmetricClamp(int x, int a, int b, int aa, int bb);
const double COEFF[5] = {1.00/16, 4.00/16, 6.00/16, 4.00/16, 1.00/16};
void downSampling(double *x, double *xx, int W, int H, int WW, int HH) {						// xx = down-sampling(x);
	for(int ii = 0;  ii < WW;  ii++) {
		for(int jj = 0;  jj < HH;  jj++) {
			xx[jj*WW+ii] = 0.00;
			for(int i = 2*ii-2;  i <= 2*ii+2;  i++) {
				for(int j = 2*jj-2;  j <= 2*jj+2;  j++) {
					int i_ = symmetricClamp(i, 0, W-1, 0, 0);
					int j_ = symmetricClamp(j, 0, H-1, 0, 0);
					xx[jj*WW+ii] += COEFF[i-2*ii+2]*COEFF[j-2*jj+2]*x[j_*W+i_];
				}
			}
		}
	}
}
void upSampling(double *x, double *xx, double *t, int W, int H, int WW, int HH, double sigma) {	// x = up-sampling(xx); t is the guiding signal.
	double sigmaSquare = sigma*sigma;
	for(int i = 0;  i < W;  i++) {
		for(int j = 0;  j < H;  j++) {
			x[j*W+i] = 0.00;
			double sum = 0.00;
			for(int ii = (i+1)/2-1;  ii <= i/2+1;  ii++) {
				for(int jj = (j+1)/2-1;  jj <= j/2+1;  jj++) {
					int ii_ = symmetricClamp(ii, 0, WW-1, 0, 1-W%2);
					int jj_ = symmetricClamp(jj, 0, HH-1, 0, 1-H%2);
					double coeff = COEFF[i-2*ii+2]*COEFF[j-2*jj+2];
					/* the joint bilateral filtering */
					sum += coeff;
					x[j*W+i] += coeff*xx[jj_*WW+ii_];
				}
			}
			x[j*W+i] /= sum;
		}
	}
}
extern inline void rangeClamp(double &x, double a, double b);
void pyramidFilter(int W, int H, double *x, double *y, double lambda, double alpha, double beta, double sigma,
				   bool clamp, bool useTable, int level, double range) {
	if(level == 0 || W <= MIN_SIZE || H <= MIN_SIZE) {
		linearFilter(W, H, x, y, beta);
		if(clamp == true) {
			for(int index = 0;  index < W*H;  index++) {
				rangeClamp(y[index], 0.00, 1.00);
			}
		}
		return;
	}

	int WW = (W+1)/2;
	int HH = (H+1)/2;
	double *xx = new double[WW*HH];
	double *yy = new double[WW*HH];
	downSampling(x, xx, W, H, WW, HH);

	pyramidFilter(WW, HH, xx, yy, lambda, alpha, beta, sigma, false, useTable, level-1, range);

	double *b = new double[W*H];
	upSampling(b, yy, x, W, H, WW, HH, sigma/2);

	MixedDomainFilter(W, H, x, y, b, lambda, alpha, beta, sigma, useTable, range);

	if(clamp == true) {
		for(int index = 0;  index < W*H;  index++) {
			rangeClamp(y[index], 0.00, 1.00);
		}
	}
	delete []xx;
	delete []yy;
	delete []b;
}


void jointPyramidFilter(int W, int H, double *x, double **ref, double *y, double lambda, double alpha, double beta, double sigma,
				   bool clamp, bool useTable, int level, double range) {
	if(level == 0 || W <= MIN_SIZE || H <= MIN_SIZE) {
		linearFilter(W, H, x, y, beta);
		if(clamp == true) {
			for(int index = 0;  index < W*H;  index++) {
				rangeClamp(y[index], 0.00, 1.00);
			}
		}
		return;
	}

	int WW = (W+1)/2;
	int HH = (H+1)/2;
	double *xx = new double[WW*HH];
	double **reff = new double*[3];
	for(int k = 0;  k < 3;  k++) {
		reff[k] = new double[WW*HH];
		downSampling(ref[k], reff[k], W, H, WW, HH);
	}
	double *yy = new double[WW*HH];
	downSampling(x, xx, W, H, WW, HH);

	jointPyramidFilter(WW, HH, xx, reff, yy, lambda, alpha, beta, sigma, false, useTable, level-1, range);

	double *b = new double[W*H];
	upSampling(b, yy, x, W, H, WW, HH, sigma/2);

	jointMixedDomainFilter(W, H, x, ref, y, b, lambda, alpha, beta, sigma, useTable, range);

	if(clamp == true) {
		for(int index = 0;  index < W*H;  index++) {
			rangeClamp(y[index], 0.00, 1.00);
		}
	}
	delete []xx;
	delete []yy;
	delete []b;
	for(int k = 0;  k < 3;  k++) {
		delete []reff[k];
	}
	delete []reff;
}
